/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS)
#include "RemoteMediaPlayerManagerProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "RemoteMediaPlayerManagerProxyMessages.h"
#include "RemoteMediaPlayerProxyConfiguration.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/MediaPlayer.h>
#include <WebCore/MediaPlayerEnums.h>
#include <WebCore/MediaPlayerIdentifier.h>
#include <WebCore/SecurityOriginData.h>
#include <wtf/HashSet.h>
#include <wtf/Vector.h>
#include <wtf/WallTime.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteMediaPlayerManagerProxy {

void GetSupportedTypes::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& types)
{
    encoder.get() << types;
    connection.sendSyncReply(WTFMove(encoder));
}

void SupportsTypeAndCodecs::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, WebCore::MediaPlayerEnums::SupportsType support)
{
    encoder.get() << support;
    connection.sendSyncReply(WTFMove(encoder));
}

void OriginsInMediaCache::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const HashSet<WebCore::SecurityOriginData>& origins)
{
    encoder.get() << origins;
    connection.sendSyncReply(WTFMove(encoder));
}

void SupportsKeySystem::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool supports)
{
    encoder.get() << supports;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteMediaPlayerManagerProxy

} // namespace Messages

namespace WebKit {

void RemoteMediaPlayerManagerProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::CreateMediaPlayer::name())
        return IPC::handleMessage<Messages::RemoteMediaPlayerManagerProxy::CreateMediaPlayer>(connection, decoder, this, &RemoteMediaPlayerManagerProxy::createMediaPlayer);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::DeleteMediaPlayer::name())
        return IPC::handleMessage<Messages::RemoteMediaPlayerManagerProxy::DeleteMediaPlayer>(connection, decoder, this, &RemoteMediaPlayerManagerProxy::deleteMediaPlayer);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::ClearMediaCache::name())
        return IPC::handleMessage<Messages::RemoteMediaPlayerManagerProxy::ClearMediaCache>(connection, decoder, this, &RemoteMediaPlayerManagerProxy::clearMediaCache);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::ClearMediaCacheForOrigins::name())
        return IPC::handleMessage<Messages::RemoteMediaPlayerManagerProxy::ClearMediaCacheForOrigins>(connection, decoder, this, &RemoteMediaPlayerManagerProxy::clearMediaCacheForOrigins);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

bool RemoteMediaPlayerManagerProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, UniqueRef<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::GetSupportedTypes::name())
        return IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerManagerProxy::GetSupportedTypes>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerManagerProxy::getSupportedTypes);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::SupportsTypeAndCodecs::name())
        return IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerManagerProxy::SupportsTypeAndCodecs>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerManagerProxy::supportsTypeAndCodecs);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::OriginsInMediaCache::name())
        return IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerManagerProxy::OriginsInMediaCache>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerManagerProxy::originsInMediaCache);
    if (decoder.messageName() == Messages::RemoteMediaPlayerManagerProxy::SupportsKeySystem::name())
        return IPC::handleMessageSynchronous<Messages::RemoteMediaPlayerManagerProxy::SupportsKeySystem>(connection, decoder, replyEncoder, this, &RemoteMediaPlayerManagerProxy::supportsKeySystem);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return false;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled synchronous message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
    return false;
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS)
