/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "RemoteSourceBufferProxy.h"

#include "ArgumentCoders.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "RemoteSourceBufferProxyMessages.h"
#include "SharedMemory.h"
#include "TrackPrivateRemoteIdentifier.h"
#include "WebCoreArgumentCoders.h"
#include <WebCore/ContentType.h>
#include <WebCore/MediaPlayer.h>
#include <WebCore/PlatformTimeRanges.h>
#include <WebCore/SourceBufferPrivate.h>
#include <utility>
#include <wtf/MediaTime.h>
#include <wtf/Vector.h>
#include <wtf/text/WTFString.h>

namespace Messages {

namespace RemoteSourceBufferProxy {

void CanSwitchToType::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, bool canSwitch)
{
    encoder.get() << canSwitch;
    connection.sendSyncReply(WTFMove(encoder));
}

void UpdateBufferedFromTrackBuffers::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::PlatformTimeRanges& buffered)
{
    encoder.get() << buffered;
    connection.sendSyncReply(WTFMove(encoder));
}

void RemoveCodedFrames::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebCore::PlatformTimeRanges&&, uint64_t&&)>&& completionHandler)
{
    std::optional<WebCore::PlatformTimeRanges> buffered;
    decoder >> buffered;
    if (!buffered) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    std::optional<uint64_t> totalTrackBufferSizeInBytes;
    decoder >> totalTrackBufferSizeInBytes;
    if (!totalTrackBufferSizeInBytes) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*buffered), WTFMove(*totalTrackBufferSizeInBytes));
}

void RemoveCodedFrames::cancelReply(CompletionHandler<void(WebCore::PlatformTimeRanges&&, uint64_t&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<WebCore::PlatformTimeRanges>::create(), IPC::AsyncReplyError<uint64_t>::create());
}

void RemoveCodedFrames::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebCore::PlatformTimeRanges& buffered, uint64_t totalTrackBufferSizeInBytes)
{
    encoder.get() << buffered;
    encoder.get() << totalTrackBufferSizeInBytes;
    connection.sendSyncReply(WTFMove(encoder));
}

void EvictCodedFrames::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, uint64_t totalTrackBufferSizeInBytes)
{
    encoder.get() << totalTrackBufferSizeInBytes;
    connection.sendSyncReply(WTFMove(encoder));
}

void BufferedSamplesForTrackId::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    std::optional<Vector<String>> samples;
    decoder >> samples;
    if (!samples) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*samples));
}

void BufferedSamplesForTrackId::cancelReply(CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<String>>::create());
}

void BufferedSamplesForTrackId::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& samples)
{
    encoder.get() << samples;
    connection.sendSyncReply(WTFMove(encoder));
}

void EnqueuedSamplesForTrackID::callReply(IPC::Decoder& decoder, CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    std::optional<Vector<String>> samples;
    decoder >> samples;
    if (!samples) {
        ASSERT_NOT_REACHED();
        cancelReply(WTFMove(completionHandler));
        return;
    }
    completionHandler(WTFMove(*samples));
}

void EnqueuedSamplesForTrackID::cancelReply(CompletionHandler<void(Vector<String>&&)>&& completionHandler)
{
    completionHandler(IPC::AsyncReplyError<Vector<String>>::create());
}

void EnqueuedSamplesForTrackID::send(UniqueRef<IPC::Encoder>&& encoder, IPC::Connection& connection, const Vector<String>& samples)
{
    encoder.get() << samples;
    connection.sendSyncReply(WTFMove(encoder));
}

} // namespace RemoteSourceBufferProxy

} // namespace Messages

namespace WebKit {

void RemoteSourceBufferProxy::didReceiveMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetActive::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetActive>(connection, decoder, this, &RemoteSourceBufferProxy::setActive);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetMode::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetMode>(connection, decoder, this, &RemoteSourceBufferProxy::setMode);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::Append::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::Append>(connection, decoder, this, &RemoteSourceBufferProxy::append);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::Abort::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::Abort>(connection, decoder, this, &RemoteSourceBufferProxy::abort);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetParserState::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetParserState>(connection, decoder, this, &RemoteSourceBufferProxy::resetParserState);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::RemovedFromMediaSource::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::RemovedFromMediaSource>(connection, decoder, this, &RemoteSourceBufferProxy::removedFromMediaSource);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetMediaSourceEnded::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetMediaSourceEnded>(connection, decoder, this, &RemoteSourceBufferProxy::setMediaSourceEnded);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetReadyState::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetReadyState>(connection, decoder, this, &RemoteSourceBufferProxy::setReadyState);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::StartChangingType::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::StartChangingType>(connection, decoder, this, &RemoteSourceBufferProxy::startChangingType);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::AddTrackBuffer::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::AddTrackBuffer>(connection, decoder, this, &RemoteSourceBufferProxy::addTrackBuffer);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetTrackBuffers::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetTrackBuffers>(connection, decoder, this, &RemoteSourceBufferProxy::resetTrackBuffers);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ClearTrackBuffers::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::ClearTrackBuffers>(connection, decoder, this, &RemoteSourceBufferProxy::clearTrackBuffers);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAllTrackBuffersNeedRandomAccess::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAllTrackBuffersNeedRandomAccess>(connection, decoder, this, &RemoteSourceBufferProxy::setAllTrackBuffersNeedRandomAccess);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::RemoveCodedFrames::name())
        return IPC::handleMessageAsync<Messages::RemoteSourceBufferProxy::RemoveCodedFrames>(connection, decoder, this, &RemoteSourceBufferProxy::removeCodedFrames);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ReenqueueMediaIfNeeded::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::ReenqueueMediaIfNeeded>(connection, decoder, this, &RemoteSourceBufferProxy::reenqueueMediaIfNeeded);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetGroupStartTimestamp::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetGroupStartTimestamp>(connection, decoder, this, &RemoteSourceBufferProxy::setGroupStartTimestamp);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetGroupStartTimestampToEndTimestamp::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetGroupStartTimestampToEndTimestamp>(connection, decoder, this, &RemoteSourceBufferProxy::setGroupStartTimestampToEndTimestamp);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetShouldGenerateTimestamps::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetShouldGenerateTimestamps>(connection, decoder, this, &RemoteSourceBufferProxy::setShouldGenerateTimestamps);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::ResetTimestampOffsetInTrackBuffers::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::ResetTimestampOffsetInTrackBuffers>(connection, decoder, this, &RemoteSourceBufferProxy::resetTimestampOffsetInTrackBuffers);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetTimestampOffset::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetTimestampOffset>(connection, decoder, this, &RemoteSourceBufferProxy::setTimestampOffset);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAppendWindowStart::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAppendWindowStart>(connection, decoder, this, &RemoteSourceBufferProxy::setAppendWindowStart);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SetAppendWindowEnd::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SetAppendWindowEnd>(connection, decoder, this, &RemoteSourceBufferProxy::setAppendWindowEnd);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::SeekToTime::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::SeekToTime>(connection, decoder, this, &RemoteSourceBufferProxy::seekToTime);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::UpdateTrackIds::name())
        return IPC::handleMessage<Messages::RemoteSourceBufferProxy::UpdateTrackIds>(connection, decoder, this, &RemoteSourceBufferProxy::updateTrackIds);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::BufferedSamplesForTrackId::name())
        return IPC::handleMessageAsync<Messages::RemoteSourceBufferProxy::BufferedSamplesForTrackId>(connection, decoder, this, &RemoteSourceBufferProxy::bufferedSamplesForTrackId);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::EnqueuedSamplesForTrackID::name())
        return IPC::handleMessageAsync<Messages::RemoteSourceBufferProxy::EnqueuedSamplesForTrackID>(connection, decoder, this, &RemoteSourceBufferProxy::enqueuedSamplesForTrackID);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled message %s to %" PRIu64, IPC::description(decoder.messageName()), decoder.destinationID());
}

bool RemoteSourceBufferProxy::didReceiveSyncMessage(IPC::Connection& connection, IPC::Decoder& decoder, UniqueRef<IPC::Encoder>& replyEncoder)
{
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::CanSwitchToType::name())
        return IPC::handleMessageSynchronous<Messages::RemoteSourceBufferProxy::CanSwitchToType>(connection, decoder, replyEncoder, this, &RemoteSourceBufferProxy::canSwitchToType);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::UpdateBufferedFromTrackBuffers::name())
        return IPC::handleMessageSynchronous<Messages::RemoteSourceBufferProxy::UpdateBufferedFromTrackBuffers>(connection, decoder, replyEncoder, this, &RemoteSourceBufferProxy::updateBufferedFromTrackBuffers);
    if (decoder.messageName() == Messages::RemoteSourceBufferProxy::EvictCodedFrames::name())
        return IPC::handleMessageSynchronous<Messages::RemoteSourceBufferProxy::EvictCodedFrames>(connection, decoder, replyEncoder, this, &RemoteSourceBufferProxy::evictCodedFrames);
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    UNUSED_PARAM(replyEncoder);
#if ENABLE(IPC_TESTING_API)
    if (connection.ignoreInvalidMessageForTesting())
        return false;
#endif // ENABLE(IPC_TESTING_API)
    ASSERT_NOT_REACHED_WITH_MESSAGE("Unhandled synchronous message %s to %" PRIu64, description(decoder.messageName()), decoder.destinationID());
    return false;
}

} // namespace WebKit

#endif // ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
